#include "linalg.h"

void invert_matrix(Array2d<Real> *inv,const Array2d<Real> &a) {
  iVector2d size=a.get_size();
  if (size(0)!=size(1)) ERRORQUIT("Cannot invert non square matrix!");
  inv->resize(size);
  int n=size(0);
  Real **linear_sys=numrec_matrix(1,n,1,n);
  for (int i=0; i<n; i++) {
    for (int j=0; j<n; j++) {
      linear_sys[1+i][1+j]=a(i,j);
    }
  }
  int *indx=numrec_ivector(1,n);
  Real *d=numrec_vector(1,n);
  ludcmp(linear_sys,n,indx,d);
  Real *rhs=numrec_vector(1,n);
  for (int col=0; col<n; col++) {
    for (int i=1; i<=n; i++) {rhs[i]=0.;}
    rhs[1+col]=1.;
    lubksb(linear_sys,n,indx,rhs);
    for (int i=0; i<n; i++) {
      (*inv)(i,col)=rhs[1+i];
    }
  }
  numrec_free_matrix(linear_sys,1,n,1,n);
  numrec_free_ivector(indx,1,n);
  numrec_free_vector(d,1,n);
  numrec_free_vector(rhs,1,n);
}

void linsolve(Array<Real> *soln, const Array2d<Real> &mat, const Array<Real> &vect) {
  int dim=vect.get_size();
  Real **pmat=numrec_matrix(1,dim,1,dim);
  Real *pvect=numrec_vector(1,dim);
  for (int i=0; i<dim; i++) {
    for (int j=0; j<dim; j++) {
      pmat[1+i][1+j]=mat(i,j);
    }
  }
  for (int i=0; i<dim; i++) {
    pvect[1+i]=vect(i);
  }
  linsolve(pmat,dim,pvect);
  soln->resize(dim);
  for (int i=0; i<dim; i++) {
    (*soln)(i)=pvect[i+1];
  }
}

void transpose_matrix(Array2d<Real> *trans,const Array2d<Real> &a) {
  int maxi=a.get_size()(1);
  int maxj=a.get_size()(0);
  trans->resize(maxi,maxj);
  for (int i=0; i<maxi; i++) {
    for (int j=0; j<maxj; j++) {
      (*trans)(i,j)=a(j,i);
    }
  }
}

void diff(Array<Real> *result, const Array<Real> &a, const Array<Real> &b) {
#ifdef DEBUG
  if (a.get_size()!=b.get_size()) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(a.get_size());
  for (int i=0; i<a.get_size(); i++) {
    (*result)(i)=a(i)-b(i);
  }
}

void diff(Array2d<Real> *result, const Array2d<Real> &a, const Array2d<Real> &b) {
#ifdef DEBUG
  if (a.get_size()!=b.get_size()) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(a.get_size());
  for (int i=0; i<a.get_size()(0); i++) {
    for (int j=0; j<a.get_size()(1); j++) {
      (*result)(i,j)=a(i,j)-b(i,j);
    }
  }
}

void sum(Array<Real> *result, const Array<Real> &a, const Array<Real> &b) {
#ifdef DEBUG
  if (a.get_size()!=b.get_size()) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(a.get_size());
  for (int i=0; i<a.get_size(); i++) {
    (*result)(i)=a(i)+b(i);
  }
}

void sum(Array2d<Real> *result, const Array2d<Real> &a, const Array2d<Real> &b) {
#ifdef DEBUG
  if (a.get_size()!=b.get_size()) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(a.get_size());
  for (int i=0; i<a.get_size()(0); i++) {
    for (int j=0; j<a.get_size()(1); j++) {
      (*result)(i,j)=a(i,j)+b(i,j);
    }
  }
}

void sum(Array2d<Complex> *result, const Array2d<Complex> &a, const Array2d<Complex> &b) {
#ifdef DEBUG
  if (a.get_size()!=b.get_size()) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(a.get_size());
  for (int i=0; i<a.get_size()(0); i++) {
    for (int j=0; j<a.get_size()(1); j++) {
      (*result)(i,j)=a(i,j)+b(i,j);
    }
  }
}

void product(Array2d<Real> *result, const Array2d<Real> &a, const Array2d<Real> &b) {
#ifdef DEBUG
  if (a.get_size()(1)!=b.get_size()(0)) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(iVector2d(a.get_size()(0),b.get_size()(1)));
  zero_array(result);
  for (int i=0; i<a.get_size()(0); i++) {
    for (int j=0; j<b.get_size()(1); j++) {
      for (int k=0; k<a.get_size()(1); k++) {
        (*result)(i,j)+=a(i,k)*b(k,j);
      }
    }
  }
}

void product_diag(Array2d<Real> *result, const Array<Real> &diag, const Array2d<Real> &b) {
#ifdef DEBUG
  if (diag.get_size()!=b.get_size()(0)) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(b.get_size());
  for (int i=0; i<b.get_size()(0); i++) {
    for (int j=0; j<b.get_size()(1); j++) {
      (*result)(i,j)=diag(i)*b(i,j);
    }
  }
}

void product_diag(Array<Real> *result, const Array<Real> &diag, const Array<Real> &b) {
#ifdef DEBUG
  if (diag.get_size()!=b.get_size()) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(b.get_size());
  for (int i=0; i<b.get_size(); i++) {
    (*result)(i)=diag(i)*b(i);
  }
}

void product(Array<Real> *result, const Array2d<Real> &a, const Array<Real> &b) {
#ifdef DEBUG
  if (a.get_size()(1)!=b.get_size()) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(a.get_size()(0));
  zero_array(result);
  for (int i=0; i<a.get_size()(0); i++) {
    for (int k=0; k<a.get_size()(1); k++) {
      (*result)(i)+=a(i,k)*b(k);
    }
  }
}

void product(Array<Real> *result, const Array<Real> &a, const Array2d<Real> &b) {
#ifdef DEBUG
  if (a.get_size()!=b.get_size()(0)) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(b.get_size()(1));
  zero_array(result);
  for (int i=0; i<b.get_size()(1); i++) {
    for (int k=0; k<a.get_size(); k++) {
      (*result)(i)+=a(k)*b(k,i);
    }
  }
}

void product(Array2d<Real> *result, const Array2d<Real> &a, Real b) {
  result->resize(a.get_size());
  for (int i=0; i<a.get_size()(0); i++) {
    for (int j=0; j<a.get_size()(1); j++) {
      (*result)(i,j)=a(i,j)*b;
    }
  }
}

void product(Array<Real> *result, const Array<Real> &a, Real b) {
  result->resize(a.get_size());
  for (int i=0; i<a.get_size(); i++) {
    (*result)(i)=a(i)*b;
  }
}

Real inner_product(const Array<Real> &a, const Array<Real> &b) {
  Real sum=0;
#ifdef DEBUG
  if (a.get_size()!=b.get_size()) ERRORQUITDUMP("Matrices not conformable");
#endif
  for (int i=0; i<a.get_size(); i++) {
    sum+=a(i)*b(i);
  }
  return sum;  
}

void inner_product(Array2d<Real> *result, const Array2d<Real> &a, const Array2d<Real> &b) {
#ifdef DEBUG
  if (a.get_size()(0)!=b.get_size()(0)) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(iVector2d(a.get_size()(1),b.get_size()(1)));
  zero_array(result);
  int nk=a.get_size()(0);
  for (int i=0; i<a.get_size()(1); i++) {
    for (int j=0; j<b.get_size()(1); j++) {
      Real s=0.;
      for (int k=0; k<nk; k++) {
        s+=a(k,i)*b(k,j);
      }
      (*result)(i,j)=s;
    }
  }
}

void inner_product(Array<Real> *result, const Array2d<Real> &a, const Array<Real> &b) {
#ifdef DEBUG
  if (a.get_size()(0)!=b.get_size()) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(a.get_size()(1));
  zero_array(result);
  for (int i=0; i<a.get_size()(1); i++) {
    for (int k=0; k<a.get_size()(0); k++) {
      (*result)(i)+=a(k,i)*b(k);
    }
  }
}

Real normalize(Array<Real> *a) {
  Real l=sqrt(inner_product(*a,*a));
  product(a,*a,1./l);
  return l;
}

void outer_product(Array2d<Real> *result, const Array2d<Real> &a, const Array2d<Real> &b) {
#ifdef DEBUG
  if (a.get_size()(1)!=b.get_size()(1)) ERRORQUITDUMP("Matrices not conformable");
#endif
  result->resize(iVector2d(a.get_size()(0),b.get_size()(0)));
  zero_array(result);
  for (int i=0; i<a.get_size()(0); i++) {
    for (int j=0; j<b.get_size()(0); j++) {
      for (int k=0; k<a.get_size()(1); k++) {
        (*result)(i,j)+=a(i,k)*b(j,k);
      }
    }
  }
}

void outer_product(Array2d<Real> *result, const Array<Real> &a, const Array<Real> &b) {
  result->resize(iVector2d(a.get_size(),b.get_size()));
  for (int i=0; i<a.get_size(); i++) {
    for (int j=0; j<b.get_size(); j++) {
      (*result)(i,j)=a(i)*b(j);
    }
  }
}

void outer_product_diag(Array<Real> *result, const Array2d<Real> &a, const Array2d<Real> &b) {
#ifdef DEBUG
  if (a.get_size()(1)!=b.get_size()(1)) ERRORQUITDUMP("Matrices not conformable");
  if (a.get_size()(0)!=b.get_size()(0)) ERRORQUITDUMP("Matrix is not square");
#endif
  result->resize(a.get_size()(0));
  zero_array(result);
  for (int i=0; i<a.get_size()(0); i++) {
    for (int k=0; k<a.get_size()(1); k++) {
      (*result)(i)+=a(i,k)*b(i,k);
    }
  }
}

Real quadratic_form(const Array2d<Real> &a, const Array<Real> &x) {
#ifdef DEBUG
  if (a.get_size()(0)!=a.get_size()(1)) ERRORQUITDUMP("Matrix is not square");
  if (a.get_size()(0)!=x.get_size())    ERRORQUITDUMP("Matrices not conformable");
#endif
  Real result=0.;
  for (int i=0; i<x.get_size(); i++) {
    for (int j=0; j<x.get_size(); j++) {
      result+=a(i,j)*x(i)*x(j);
    }
  }
  return result;
}

Real quadratic_form(const Array2d<Complex> &a, const Array<Complex> &x) {
#ifdef DEBUG
  if (a.get_size()(0)!=a.get_size()(1)) ERRORQUITDUMP("Matrix is not square");
  if (a.get_size()(0)!=x.get_size())    ERRORQUITDUMP("Matrices not conformable");
#endif
  Real result=0.;
  for (int i=0; i<x.get_size(); i++) {
    for (int j=0; j<x.get_size(); j++) {
      result+=real((a(i,j)*conj(x(i))*x(j)));
    }
  }
  return result;
}

Real trace(const Array2d<Real> &a) {
#ifdef DEBUG
  if (a.get_size()(0)!=a.get_size()(1)) ERRORQUITDUMP("Cannot take trace of nonsquare matrix");
#endif
  Real tr=0;
  for (int i=0; i<a.get_size()(0); i++) {
    tr+=a(i,i);
  }
  return tr;
}

void set_identity(Array2d<Real> *result, int n) {
  result->resize(iVector2d(n,n));
  zero_array(result);
  for (int i=0; i<n; i++) {
    (*result)(i,i)=1.;
  }
}

void extract_columns(Array2d<Real> *pa, const Array2d<Real> &b, const Array<int> &cols, int nb_cols) {
  if (nb_cols==-1) {nb_cols=cols.get_size();}
  pa->resize(iVector2d(b.get_size()(0),nb_cols));
  for (int i=0; i<nb_cols; i++) {
    if (cols(i)>=0) {
      for (int j=0; j<b.get_size()(0); j++) {
        (*pa)(j,i)=b(j,cols(i));
      }
    }
  }
}

void extract_columns(Array2d<Real> *pa, const Array2d<Real> &b, const LinkedList<int> &cols_list, int nb_cols) {
  Array<int> cols;
  LinkedList_to_Array(&cols,cols_list);
  extract_columns(pa,b,cols,nb_cols);
}

void extract_column(Array<Real> *pa, const Array2d<Real> &b, int col) {
  pa->resize(b.get_size()(0));
  for (int i=0; i<b.get_size()(0); i++) {
    (*pa)(i)=b(i,col);
  }
}

void extract_row(Array<Real> *pa, const Array2d<Real> &b, int row) {
  pa->resize(b.get_size()(1));
  for (int i=0; i<b.get_size()(1); i++) {
    (*pa)(i)=b(row,i);
  }
}

void extract_elements(Array<Real> *pa, const Array<Real> &b, const Array<int> &cols, int nb_cols) {
  if (nb_cols==-1) {nb_cols=cols.get_size();}
  pa->resize(nb_cols);
  for (int i=0; i<nb_cols; i++) {
    if (cols(i)>=0) {
      (*pa)(i)=b(cols(i));
    }
  }
}

void extract_elements(Array<Real> *pa,  const Array<Real> &b, int first, int lastp1) {
  pa->resize(lastp1-first);
  int i=0;
  for (int j=first; j<lastp1; j++) {
    (*pa)(i)=b(j);
    i++;
  }
}

void to_real(Array<Real> *p_r, const Array<int> &i) {
  p_r->resize(i.get_size());
  for (int j=0; j<i.get_size(); j++) {(*p_r)(j)=(Real)i(j);}
}

void diagonalize_symmetric_matrix(Array<Real> *p_lambda, Array2d<Real> *p_vect, const Array2d<Real> &mat) {
#ifdef DEBUG
	if (mat.get_size()(0)!=mat.get_size()(1)) ERRORQUIT("Matrix not square in diagonalize_symmetric_matrix");
#endif
	int n=mat.get_size()(0);
	int do_vect=(p_vect!=NULL);
	Real **a=numrec_matrix(1,n,1,n);
	for (int i=0; i<n; i++) {
		for (int j=0; j<n; j++) {
			a[i+1][j+1]=mat(i,j);
		}
	}
	Real *d=numrec_vector(1,n);
	Real *e=numrec_vector(1,n);
	tred2(a,n,d,e,do_vect);
	tqli(d,e,n,a,do_vect);
	p_lambda->resize(n);
	for (int i=0; i<n; i++) {
	  (*p_lambda)(i)=d[i+1];
	}
	if (do_vect) {
		p_vect->resize(iVector2d(n,n));
		for (int i=0; i<n; i++) {
			for (int j=0; j<n; j++) {
				(*p_vect)(i,j)=a[i+1][j+1];
			}
		}
	}
}

void diagonalize_symmetric_matrix(Array<Real> *p_lambda, Array2d<Complex> *p_vect, const Array2d<Complex> &mat) {
#ifdef DEBUG
	if (mat.get_size()(0)!=mat.get_size()(1)) ERRORQUIT("Matrix not square in diagonalize_symmetric_matrix");
#endif
	int n=mat.get_size()(0);
	int do_vect=(p_vect!=NULL);
	Real **a=numrec_matrix(1,2*n,1,2*n);
	for (int i=0; i<n; i++) {
		for (int j=0; j<n; j++) {
			a[i+1][j+1]=real(mat(i,j));
			a[n+i+1][n+j+1]=real(mat(i,j));
		}
	}
	for (int i=0; i<n; i++) {
		for (int j=0; j<n; j++) {
			a[n+i+1][j+1]=imag(mat(i,j));
			a[i+1][n+j+1]=imag(mat(j,i));
		}
	}
	Real *d=numrec_vector(1,2*n);
	Real *e=numrec_vector(1,2*n);
	tred2(a,2*n,d,e,do_vect);
	tqli(d,e,2*n,a,do_vect);
	p_lambda->resize(n);
	if (do_vect) {
	  p_vect->resize(iVector2d(n,n));
	}
	Array<int> to_omit(2*n);
	zero_array(&to_omit);
	int j=0;
	for (int i=0; i<n; i++) {
	  while (to_omit(j)==1) j++;
	  (*p_lambda)(i)=d[j+1];
	  if (do_vect) {
	    for (int k=0; k<n; k++) {
	      (*p_vect)(k,i)=a[k+1][j+1];
	    }
	  }
	  to_omit(j)=1;
	  int k;
	  int best_k=-1;
	  Real best_d=MAXFLOAT;
	  for (k=j+1; k<2*n; k++) {
	    if (to_omit(k)!=1 && fabs(d[k+1]-d[j+1])<best_d) {
	      best_d=fabs(d[k+1]-d[j+1]);
	      best_k=k;
	    }
	  }
	  to_omit(best_k)=1;
	}
}

