#include "linsolve.h"

#define NR_END 1
#define FREE_ARG char*

Real *numrec_vector(long nl, long nh)
/* allocate a Real vector with subscript range v[nl..nh] */
{
	Real *v;

	v=(Real *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(Real)));
	if (!v) ERRORQUIT("allocation failure in vector()");
	return v-nl+NR_END;
}

int *numrec_ivector(long nl, long nh)
/* allocate an int vector with subscript range v[nl..nh] */
{
	int *v;

	v=(int *)malloc((size_t) ((nh-nl+1+NR_END)*sizeof(int)));
	if (!v) ERRORQUIT("allocation failure in ivector()");
	return v-nl+NR_END;
}

void numrec_free_vector(Real *v, long nl, long nh)
/* free a Real vector allocated with vector() */
{
	free((FREE_ARG) (v+nl-NR_END));
}

void numrec_free_ivector(int *v, long nl, long nh)
/* free an int vector allocated with ivector() */
{
	free((FREE_ARG) (v+nl-NR_END));
}

Real **numrec_matrix(long nrl, long nrh, long ncl, long nch)
/* allocate a Real matrix with subscript range m[nrl..nrh][ncl..nch] */
{
	long i, nrow=nrh-nrl+1,ncol=nch-ncl+1;
	Real **m;

	/* allocate pointers to rows */
	m=(Real **) malloc((size_t)((nrow+NR_END)*sizeof(Real*)));
	if (!m) ERRORQUIT("allocation failure 1 in matrix()");
	m += NR_END;
	m -= nrl;

	/* allocate rows and set pointers to them */
	m[nrl]=(Real *) malloc((size_t)((nrow*ncol+NR_END)*sizeof(Real)));
	if (!m[nrl]) ERRORQUIT("allocation failure 2 in matrix()");
	m[nrl] += NR_END;
	m[nrl] -= ncl;

	for (i=nrl+1;i<=nrh;i++) {m[i]=m[i-1]+ncol;}

	/* return pointer to array of pointers to rows */
	return m;
}

void numrec_free_matrix(Real **m, long nrl, long nrh, long ncl, long nch)
/* free a Real matrix allocated by matrix() */
{
	free((FREE_ARG) (m[nrl]+ncl-NR_END));
	free((FREE_ARG) (m+nrl-NR_END));
}

#define TINY 1.0e-20;

void ludcmp(Real **a, int n, int *indx, Real *d) {
	int i,imax,j,k;
	Real big,dum,sum,temp;
	Real *vv;

	vv=numrec_vector(1,n);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++) {
			if ((temp=fabs(a[i][j])) > big) big=temp;
		}
		if (big == 0.0)
		  ERRORQUIT("Singular matrix in routine LUDCMP");
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) {sum -= a[i][k]*a[k][j];}
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++) {
				sum -= a[i][k]*a[k][j];
			}
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) {a[i][j] *= dum;}
		}
	}
	numrec_free_vector(vv,1,n);
}

#undef TINY

void lubksb(Real **a, int n, int *indx, Real *b) {
	int i,ii=0,ip,j;
	Real sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
		  for (j=ii;j<=i-1;j++) {sum -= a[i][j]*b[j];}
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) {sum -= a[i][j]*b[j];}
		b[i]=sum/a[i][i];
	}
}

void linsolve(Real **a, int n, Real *rhs) {
  int *indx=numrec_ivector(1,n);
  Real *d=numrec_vector(1,n);
  ludcmp(a,n,indx,d);
  lubksb(a,n,indx,rhs);
  numrec_free_ivector(indx,1,n);
  numrec_free_vector(d,1,n);
}

/* eigenvalue problems */

void tred2(Real **a, int n, Real *d, Real *e, int do_vect) {
	int l,k,j,i;
	Real scale,hh,h,g,f;

	for (i=n;i>=2;i--) {
		l=i-1;
		h=scale=0.0;
		if (l > 1) {
		  for (k=1;k<=l;k++) {
				scale += fabs(a[i][k]);
		  }
			if (scale == 0.0)
				e[i]=a[i][l];
			else {
				for (k=1;k<=l;k++) {
					a[i][k] /= scale;
					h += a[i][k]*a[i][k];
				}
				f=a[i][l];
				g = f>0 ? -sqrt(h) : sqrt(h);
				e[i]=scale*g;
				h -= f*g;
				a[i][l]=f-g;
				f=0.0;
				for (j=1;j<=l;j++) {
				/* Next statement can be omitted if eigenvectors not wanted */
					a[j][i]=a[i][j]/h;
					g=0.0;
					for (k=1;k<=j;k++) {
						g += a[j][k]*a[i][k];
					}
					for (k=j+1;k<=l;k++) {
						g += a[k][j]*a[i][k];
					}
					e[j]=g/h;
					f += e[j]*a[i][j];
				}
				hh=f/(h+h);
				for (j=1;j<=l;j++) {
					f=a[i][j];
					e[j]=g=e[j]-hh*f;
					for (k=1;k<=j;k++) {
						a[j][k] -= (f*e[k]+g*a[i][k]);
					}
				}
			}
		} else
			e[i]=a[i][l];
		d[i]=h;
	}
	/* Next statement can be omitted if eigenvectors not wanted */
	d[1]=0.0;
	e[1]=0.0;
	/* Contents of this loop can be omitted if eigenvectors not
			wanted except for statement d[i]=a[i][i]; */
	if (do_vect) {
		for (i=1;i<=n;i++) {
			l=i-1;
			if (d[i]) {
				for (j=1;j<=l;j++) {
					g=0.0;
					for (k=1;k<=l;k++) {
						g += a[i][k]*a[k][j];
					}
					for (k=1;k<=l;k++) {
						a[k][j] -= g*a[k][i];
					}
				}
			}
			d[i]=a[i][i];
			a[i][i]=1.0;
			for (j=1;j<=l;j++) {a[j][i]=a[i][j]=0.0;}
		}
	}
	else {
		for (i=1;i<=n;i++) {
			d[i]=a[i][i];
		}
	}
}

#define SIGN(a,b) ((b)<0 ? -fabs(a) : fabs(a))

void tqli(Real *d, Real*e, int n, Real **z, int do_vect) {
	int m,l,iter,i,k;
	float s,r,p,g,f,dd,c,b;

	for (i=2;i<=n;i++) {e[i-1]=e[i];}
	e[n]=0.0;
	for (l=1;l<=n;l++) {
		iter=0;
		do {
			for (m=l;m<=n-1;m++) {
				dd=fabs(d[m])+fabs(d[m+1]);
				if (fabs(e[m])+dd == dd) break;
			}
			if (m != l) {
				//if (iter++ == 30) ERRORQUIT("Too many iterations in TQLI");
				g=(d[l+1]-d[l])/(2.0*e[l]);
				r=sqrt((g*g)+1.0);
				g=d[m]-d[l]+e[l]/(g+SIGN(r,g));
				s=c=1.0;
				p=0.0;
				for (i=m-1;i>=l;i--) {
					f=s*e[i];
					b=c*e[i];
					if (fabs(f) >= fabs(g)) {
						c=g/f;
						r=sqrt((c*c)+1.0);
						e[i+1]=f*r;
						c *= (s=1.0/r);
					} else {
						s=f/g;
						r=sqrt((s*s)+1.0);
						e[i+1]=g*r;
						s *= (c=1.0/r);
					}
					g=d[i+1]-p;
					r=(d[i]-g)*s+2.0*c*b;
					p=s*r;
					d[i+1]=g+p;
					g=c*r-b;
					/* Next loop can be omitted if eigenvectors not wanted */
					if (do_vect) {
						for (k=1;k<=n;k++) {
							f=z[k][i+1];
							z[k][i+1]=s*z[k][i]+c*f;
							z[k][i]=c*z[k][i]-s*f;
						}
					}
				}
				d[l]=d[l]-p;
				e[l]=g;
				e[m]=0.0;
			}
		} while (m != l);
	}
}

