#include "mmclib.h"

typedef int *pint;
typedef int **ppint;
typedef Real *pReal;
typedef Real **ppReal;
typedef Real ***pppReal;
typedef Complex *pComplex;

MultiMonteCarlo::MultiMonteCarlo(const Structure &_lattice, const Array<Array<int> > &_site_type_list, const iVector3d &_supercell, 
		const SpaceGroup &space_group, const LinkedList<MultiCluster> &cluster_list, 
		const Array<Array<Array<Real> > > &_corrfunc) :
               lattice(_lattice), site_type_list(_site_type_list), supercell(_supercell), cur_rho(), cur_conc(), mu() {
    Real max_clus_len=0.;
    {
      LinkedListIterator<MultiCluster> c(cluster_list);
      for ( ; c; c++) {
        Real l=get_cluster_length(c->clus);
        if (l>max_clus_len) max_clus_len=l;
      }
    }
    margin=find_sphere_bounding_box(lattice.cell,max_clus_len);
    iVector3d supercell_orig=supercell;
    for (int mult=1; 1 ; mult++) {
      supercell=mult*supercell_orig;
      int i;
      for (i=0; i<3; i++) {
        if (supercell(i) < margin(i)) break;
      }
      if (i==3) break;
    }
    total_box=supercell+2*margin;
    int spin_total_size=total_box(0)*total_box(1)*total_box(2)*lattice.atom_pos.get_size();
    spin=new SPIN_TYPE[spin_total_size];
    for (int i=0; i<spin_total_size; i++) {
      spin[i]=0;
    }
    cur_energy=0;
    int spin_size=supercell(0)*supercell(1)*supercell(2)*lattice.atom_pos.get_size();
    spin_orig=new SPIN_TYPE[spin_size];
    for (int i=0; i<spin_size; i++) {
      spin_orig[i]=0;
    }
    cur_disorder_param=0.;

    rspin_size=(Real)spin_size;

    site_in_cell=lattice.atom_pos.get_size();
    nb_spin_val=new int[site_in_cell];
    for (int i=0; i<site_in_cell; i++) {
	nb_spin_val[i]=site_type_list(lattice.atom_type(i)).get_size();
    }
    nb_clusters=new int[site_in_cell];
    nb_clusters=new int[site_in_cell];
    cluster_size=new pint[site_in_cell];
    site_offset=new ppint[site_in_cell];
    spin_val_clus=new pppReal[site_in_cell];
    eci=new pReal[site_in_cell];
    which_cluster=new pint[site_in_cell];
    total_clusters=cluster_list.get_size();
    cur_rho.resize(total_clusters);
    zero_array(&cur_rho);
    pcur_rho=cur_rho.get_buf();
    new_rho=new Real[total_clusters];
    rcluster_mult_per_atom=new Real[total_clusters];
    {
      LinkedListIterator<MultiCluster> c(cluster_list);
      for (int i=0; c; c++,i++) {
	rcluster_mult_per_atom[i]=calc_multiplicity(*c,lattice.cell,space_group.point_op,space_group.trans)/(Real)lattice.atom_pos.get_size();
      }
    }

    E_ref=0.;
    which_is_empty=-1;
    nb_point=0;
    {
      LinkedListIterator<MultiCluster> c(cluster_list);
      int cluster_index=0;
      for (; c; c++, cluster_index++) {
	  if (c->clus.get_size()==0) {which_is_empty=cluster_index;}
	  if (c->clus.get_size()==1) {nb_point++;}
	}
      which_is_point=new int[nb_point];
      point_mult=new Real[nb_point];
	c.init(cluster_list);
      cluster_index=0;
      int point_index=0;
      for (; c; c++, cluster_index++) {
            if (c->clus.get_size()==1) {
              which_is_point[point_index]=cluster_index;
              point_mult[point_index]=rcluster_mult_per_atom[cluster_index];
              point_index++;
            }
	}
    }
    cur_conc.resize(nb_point);
    mu.resize(nb_point);
    zero_array(&mu);
    pmu=mu.get_buf();

    rMatrix3d inv_cell=!lattice.cell;
    for (int s=0; s<lattice.atom_pos.get_size(); s++) {
      LinkedList<MultiCluster> long_cluster_list;
      LinkedList<int> long_cluster_i_list;
      LinkedListIterator<MultiCluster> c(cluster_list);
      int cluster_index=0;
      
      for ( ; c; c++,cluster_index++) {
	if (c->clus.get_size()!=0) {
            Array<MultiCluster> clusters;
            find_equivalent_clusters(&clusters, *c,space_group.cell,space_group.point_op,space_group.trans);
            for (int ec=0; ec<clusters.get_size(); ec++) {
	      for (int center_atom=0; center_atom<clusters(ec).clus.get_size(); center_atom++) {
		rVector3d lat_shift=lattice.atom_pos(s)-clusters(ec).clus(center_atom);
		if (is_int(inv_cell*lat_shift)) {
                  MultiCluster *to_add=new MultiCluster(clusters(ec).clus.get_size());
                  to_add->clus(0)=clusters(ec).clus(center_atom)+lat_shift;
		  to_add->site_type(0)=clusters(ec).site_type(center_atom);
		  to_add->func(0)=clusters(ec).func(center_atom);
                  int j=1;
                  for (int i=0; i<clusters(ec).clus.get_size(); i++) {
                    if (i!=center_atom) {
                      to_add->clus(j)=clusters(ec).clus(i)+lat_shift;
                      to_add->site_type(j)=clusters(ec).site_type(i);
                      to_add->func(j)=clusters(ec).func(i);
                      j++;
                    }
                  }
                  long_cluster_list << to_add;
                  long_cluster_i_list << new int(cluster_index);
                }
              }
            }
        }
      }
      nb_clusters[s]=long_cluster_list.get_size();
      cluster_size[s]=new int[nb_clusters[s]];
      eci[s]=new Real[nb_clusters[s]];
      which_cluster[s]=new int[nb_clusters[s]];
      site_offset[s]=new pint[nb_clusters[s]];
      spin_val_clus[s]=new ppReal[nb_clusters[s]];
      c.init(long_cluster_list);
      LinkedListIterator<int> c_i(long_cluster_i_list);
      for (int ic=0; ic<nb_clusters[s]; ic++, c++, c_i++) {
        cluster_size[s][ic]=c->clus.get_size();
	eci[s][ic]=0.;
        which_cluster[s][ic]=*c_i;
        site_offset[s][ic]=new int[c->clus.get_size()];
        spin_val_clus[s][ic]=new pReal[c->clus.get_size()];
        for (int j=0; j<c->clus.get_size(); j++) {
          int offset_in_cell=which_atom(lattice.atom_pos,c->clus(j),inv_cell);
          iVector3d offset_cell=to_int(inv_cell*(c->clus(j)-lattice.atom_pos(offset_in_cell)));
          site_offset[s][ic][j]=((offset_cell(0)*total_box(1) + offset_cell(1))*total_box(2) + offset_cell(2))*site_in_cell + offset_in_cell - s;
	  spin_val_clus[s][ic][j]=new Real[nb_spin_val[offset_in_cell]];
	  for (int l=0; l<nb_spin_val[offset_in_cell]; l++) {
            spin_val_clus[s][ic][j][l]=_corrfunc(c->site_type(j))(c->func(j))(l);
          }
        }
      }
    }
  }

MultiMonteCarlo::~MultiMonteCarlo(void) {
  //to write: delete all dynamically allocated arrays;
}

void MultiMonteCarlo::set_eci(const Array<Real> &new_eci) {
  if (which_is_empty>=0) {
    E_ref=new_eci(which_is_empty)/lattice.atom_pos.get_size();
  } else {
    E_ref=0.;
  }
  for (int s=0; s<site_in_cell; s++) {
    for (int c=0; c<nb_clusters[s]; c++) {
      eci[s][c]=new_eci(which_cluster[s][c]);
    }
  }
  calc_from_scratch();
}

void MultiMonteCarlo::init_random(const Array<Array<Real> > &conc) {
    Array<Array<Real> > sconc;
    sconc=conc;
    for (int i=0; i<sconc.get_size(); i++) {
      Real s=0.;
      for (int j=0; j<sconc(i).get_size(); j++) {
         s+=conc(i)(j);
         sconc(i)(j)=s;
      }
      for (int j=0; j<sconc(i).get_size(); j++) {
         sconc(i)(j)/=s;
      }     
    }
    iMatrix3d msupercell;
    msupercell.diag(supercell);
    BoundingBox<int,3> bb(iVector3d(0,0,0),total_box-iVector3d(1,1,1));
    MultiDimIterator<iVector3d> cell(supercell);
    for ( ; cell; cell++) {
      for (int offset_in_cell=0; offset_in_cell<site_in_cell; offset_in_cell++) {
        Real r=uniform01();
        int the_spin=0;
        for (; the_spin<sconc(offset_in_cell).get_size()-1; the_spin++) {
          if (r<sconc(offset_in_cell)(the_spin)) break;
        }
        MultiDimIterator<iVector3d> image(iVector3d(-1,-1,-1),iVector3d(1,1,1));
        for ( ; image; image++) {
          iVector3d image_cell=(iVector3d &)cell+msupercell*(iVector3d &)image+margin;
          if (bb.is_in(image_cell)) {
            int offset=((image_cell(0)*total_box(1) + image_cell(1))*total_box(2) + image_cell(2))*site_in_cell + offset_in_cell;
            spin[offset]=the_spin;
          }
        }
      }
    }
    int spin_size=supercell(0)*supercell(1)*supercell(2)*site_in_cell;
    for (int i=0; i<spin_size; i++) {
      spin_orig[i]=0;
    }
    calc_from_scratch();
    // extension_calc_from_scratch();
  }

void MultiMonteCarlo::init_structure(const Structure &str) {
  rMatrix3d inv_cell=!str.cell;
  iMatrix3d msupercell;
  msupercell.diag(supercell);
  BoundingBox<int,3> bb(iVector3d(0,0,0),total_box-iVector3d(1,1,1));
  MultiDimIterator<iVector3d> cell(supercell);
  for (; cell; cell++) {
    rVector3d rcell=lattice.cell*to_real((iVector3d &)cell);
    for (int s=0; s<site_in_cell; s++) {
      SPIN_TYPE newspin=(SPIN_TYPE)str.atom_type(which_atom(str.atom_pos,rcell+lattice.atom_pos(s),inv_cell));
      MultiDimIterator<iVector3d> image(iVector3d(-1,-1,-1),iVector3d(1,1,1));
      for ( ; image; image++) {
        iVector3d image_cell=cell+msupercell*(iVector3d &)image+margin;
        if (bb.is_in(image_cell)) {
          int offset=((image_cell(0)*total_box(1) + image_cell(1))*total_box(2) + image_cell(2))*site_in_cell + s;
          spin[offset]=newspin;
        }
      }
      iVector3d thecell=(iVector3d &)cell;
      int offset=((thecell(0)*supercell(1) + thecell(1))*supercell(2) + thecell(2))*site_in_cell + s;
      spin_orig[offset]=newspin;
    }
  }
  calc_from_scratch();
//  extension_calc_from_scratch();
}

/*
void MultiMonteCarlo::set_concentration(Real concentration) {
    int nbspin=(int)(rspin_size*(concentration-get_cur_concentration())/2);
    int toflip=(nbspin>0 ? -1 : 1);
    nbspin=abs(nbspin);
    int cell[3],mcell[3],mcellscan[3],incell,offset;
    
    for (int s=0; s<nbspin; s++) {
      do {
	for (int i=0; i<3; i++) {
	  cell[i]=random(supercell(i));
	  mcell[i]=cell[i]+margin(i);
	}
	incell=random(site_in_cell);
	offset=((mcell[0]*total_box(1) + mcell[1])*total_box(2) + mcell[2])*site_in_cell + incell;
      } while (spin[offset]!=toflip);
      for (int i=0; i<3; i++) {
        if (mcell[i]>=supercell(i)) {mcell[i]-=supercell(i);}
      }
      for (mcellscan[0]=mcell[0]; mcellscan[0]<total_box(0); mcellscan[0]+=supercell(0)) {
        for (mcellscan[1]=mcell[1]; mcellscan[1]<total_box(1); mcellscan[1]+=supercell(1)) {
          for (mcellscan[2]=mcell[2]; mcellscan[2]<total_box(2); mcellscan[2]+=supercell(2)) {
            spin[((mcellscan[0]*total_box(1) + mcellscan[1])*total_box(2) + mcellscan[2])*site_in_cell + incell]*=-1;
          }
        }
      }
      offset=((cell[0]*supercell(1) + cell[1])*supercell(2) + cell[2])*site_in_cell + incell;
      spin_changed[offset]^=1;
    }
    calc_from_scratch();
    extension_calc_from_scratch();
}
*/

void MultiMonteCarlo::calc_from_scratch(void) {
    cur_disorder_param=0.;
    cur_energy=0;
    for (int i=0; i<total_clusters; i++) {cur_rho(i)=0.;}
    MultiDimIterator<iVector3d> cur_cell(supercell);
    for ( ; cur_cell; cur_cell++) {
      iVector3d m_cur_cell=(iVector3d &)cur_cell+margin;
      for (int s=0; s<site_in_cell; s++) {
        int moffset=((m_cur_cell(0)*total_box(1) + m_cur_cell(1))*total_box(2) + m_cur_cell(2))*site_in_cell + s;
        iVector3d cell=cur_cell;
        int offset=((cell(0)*supercell(1) + cell(1))*supercell(2) + cell(2))*site_in_cell + s;
        cur_disorder_param+=(spin[moffset]!=spin_orig[offset]);
        for (int c=0; c<nb_clusters[s]; c++) {
          Real rho=1;
          for (int i=0; i<cluster_size[s][c]; i++) {
            rho*=spin_val_clus[s][c][i][spin[moffset+site_offset[s][c][i]]];
          }
	  rho/=(Real)(cluster_size[s][c]);
          cur_energy+=rho*eci[s][c];
	  pcur_rho[which_cluster[s][c]]+=rho/rcluster_mult_per_atom[which_cluster[s][c]];
        }
      }
    }
    cur_disorder_param/=rspin_size;
    cur_energy/=rspin_size;
    product(&cur_rho,cur_rho,1./rspin_size);
    if (which_is_empty>=0) {cur_rho(which_is_empty)=1.;}
    for (int i=0; i<nb_point; i++) {
      cur_energy-=mu(i)*cur_rho(which_is_point[i])*point_mult[i];
    }
  }

const Array<Real> & MultiMonteCarlo::get_cur_concentration(void) {
  for (int i=0; i<nb_point; i++) {
    cur_conc(i)=cur_rho(which_is_point[i])*point_mult[i];
  }
  return cur_conc;
}

void MultiMonteCarlo::set_T_mu(Real _T, const Array<Real> &_mu) {
    T=_T;
    for (int i=0; i<nb_point; i++) {
      cur_energy-=(_mu(i)-mu(i))*cur_rho(which_is_point[i])*point_mult[i];
    }
    mu=_mu;
}

void MultiMonteCarlo::spin_flip(void) {
    int i;
    int cell[3],mcell[3],mcellscan[3],incell,offset;
    int oldspin,newspin;
    Real denergy,dconc;
    int cluster_count,site_count;
    int **pcluster;
    Real ***pppspin_val_clus,**ppspin_val_clus;
    int *pwhich_cluster;
    int *poffset,*psize;
    Real *peci, *prho, *prho2, *prcluster_mult_per_atom;
    Real rho;
    int accept;
    for (i=0; i<3; i++) {
      cell[i]=random(supercell(i));
      mcell[i]=cell[i]+margin(i);
    }
    incell=random(site_in_cell);
    offset=((mcell[0]*total_box(1) + mcell[1])*total_box(2) + mcell[2])*site_in_cell + incell;
    oldspin=spin[offset];
    newspin=(oldspin+1+random(nb_spin_val[incell]-1)) % nb_spin_val[incell];
    denergy=0.;
    for (i=total_clusters, prho=new_rho; i>0; i--, prho++) {
      *prho=0.;
    }
    for (cluster_count=nb_clusters[incell], pcluster=site_offset[incell], pppspin_val_clus=spin_val_clus[incell], psize=cluster_size[incell], pwhich_cluster=which_cluster[incell], peci=eci[incell]; cluster_count>0; cluster_count--, pcluster++, pppspin_val_clus++, psize++, pwhich_cluster++, peci++) {
      poffset=*pcluster;
      ppspin_val_clus=*pppspin_val_clus;
      rho=(*ppspin_val_clus)[newspin]-(*ppspin_val_clus)[oldspin];
      poffset++;
      ppspin_val_clus++;
      for (site_count=(*psize)-1; site_count>0; site_count--, poffset++, ppspin_val_clus++) {
        rho*=(*ppspin_val_clus)[spin[offset+(*poffset)]];
      }
      new_rho[*pwhich_cluster]+=rho;
      denergy+=(*peci)*rho;
    }
    for (i=0; i<nb_point; i++) {
      denergy-=pmu[i]*new_rho[which_is_point[i]]/point_mult[i];
    }
    // denergy+=extension_spin_flip_energy(cell,incell,-spin[offset]);
    accept=0;
    if (denergy<0) {
      accept=1;
    }
    else {
      if (uniform01()<exp(-denergy/T)) accept=1;
    }
    if (accept) {
      cur_energy+=denergy/rspin_size;
      for (i=0, prho=new_rho, prho2=pcur_rho, prcluster_mult_per_atom=rcluster_mult_per_atom; i<total_clusters; i++, prho++, prho2++, prcluster_mult_per_atom++) {
         *prho2+=(*prho)/(*prcluster_mult_per_atom)/rspin_size;
      }

      for (i=0; i<3; i++) {
        if (mcell[i]>=supercell(i)) {mcell[i]-=supercell(i);}
      }
      for (mcellscan[0]=mcell[0]; mcellscan[0]<total_box(0); mcellscan[0]+=supercell(0)) {
        for (mcellscan[1]=mcell[1]; mcellscan[1]<total_box(1); mcellscan[1]+=supercell(1)) {
          for (mcellscan[2]=mcell[2]; mcellscan[2]<total_box(2); mcellscan[2]+=supercell(2)) {
            spin[((mcellscan[0]*total_box(1) + mcellscan[1])*total_box(2) + mcellscan[2])*site_in_cell + incell]=newspin;
          }
        }
      }
      offset=((cell[0]*supercell(1) + cell[1])*supercell(2) + cell[2])*site_in_cell + incell;
      cur_disorder_param+=((newspin!=spin_orig[offset])-(oldspin!=spin_orig[offset]))/rspin_size;
    }
    //extension_update_spin_flip(cell,incell,spin[offset]);
  }

/*
void MultiMonteCarlo::spin_double_flip(void) {
    int i,f;
    int array_cell[6],array_mcell[6],array_bmcell[6];
    int *cell[2]={array_cell,array_cell+3};
    int *mcell[2]={array_mcell,array_mcell+3};
    int *bmcell[2]={array_bmcell,array_bmcell+3};
    int mcellscan[3],incell[2],offset[2],nomoffset;
    Real denergy;
    int cluster_count,site_count;
    int **pcluster;
    int *pwhich_cluster;
    int *poffset,*psize;
    Real *peci;
    int rho;
    int accept;

    offset[1]=0;
    for (f=0; f<2; f++) {
      do {
        for (i=0; i<3; i++) {
          cell[f][i]=random(supercell(i));
          mcell[f][i]=cell[f][i]+margin(i);
        }
        incell[f]=random(site_in_cell);
        offset[f]=((mcell[f][0]*total_box(1) + mcell[f][1])*total_box(2) + mcell[f][2])*site_in_cell + incell[f];
      } while (f==1 && spin[offset[0]]==spin[offset[1]]);
      for (i=0; i<3; i++) {
	bmcell[f][i]=mcell[f][i];
	if (bmcell[f][i]>=supercell(i)) {bmcell[f][i]-=supercell(i);}
      }
    }

    denergy=0.;
    for (f=0; f<2; f++) {
      for (cluster_count=nb_clusters[incell[f]], pcluster=site_offset[incell[f]], psize=cluster_size[incell[f]], peci=eci[incell[f]]; cluster_count>0; cluster_count--, pcluster++, psize++, peci++) {
        rho=spin[offset[f]];
        for (site_count=*psize, poffset=*pcluster; site_count>0; site_count--, poffset++) {
          rho*=spin[offset[f]+(*poffset)];
        }
        denergy+=-2.*(*peci)*(Real)rho;
      }
      denergy+=extension_spin_flip_energy(cell[f],incell[f],-spin[offset[f]]);
      if (f==0) {
        for (mcellscan[0]=bmcell[0][0]; mcellscan[0]<total_box(0); mcellscan[0]+=supercell(0)) {
          for (mcellscan[1]=bmcell[0][1]; mcellscan[1]<total_box(1); mcellscan[1]+=supercell(1)) {
            for (mcellscan[2]=bmcell[0][2]; mcellscan[2]<total_box(2); mcellscan[2]+=supercell(2)) {
              spin[((mcellscan[0]*total_box(1) + mcellscan[1])*total_box(2) + mcellscan[2])*site_in_cell + incell[0]]*=-1;
            }
          }
        }
        extension_update_spin_flip(cell[0],incell[0],spin[offset[0]]);
      }
    }

    accept=0;
    if (denergy<0) {
      accept=1;
    }
    else {
      if (uniform01()<exp(-denergy/T)) accept=1;
    }
 
    if (accept) {
      for (f=0; f<2; f++) {
        if (f==1) {
          for (mcellscan[0]=bmcell[f][0]; mcellscan[0]<total_box(0); mcellscan[0]+=supercell(0)) {
            for (mcellscan[1]=bmcell[f][1]; mcellscan[1]<total_box(1); mcellscan[1]+=supercell(1)) {
              for (mcellscan[2]=bmcell[f][2]; mcellscan[2]<total_box(2); mcellscan[2]+=supercell(2)) {
                spin[((mcellscan[0]*total_box(1) + mcellscan[1])*total_box(2) + mcellscan[2])*site_in_cell + incell[f]]*=-1;
              }
            }
          }
        }
        for (cluster_count=nb_clusters[incell[f]], pcluster=site_offset[incell[f]], psize=cluster_size[incell[f]], pwhich_cluster=which_cluster[incell[f]]; cluster_count>0; cluster_count--, pcluster++, psize++, pwhich_cluster++) {
          rho=-spin[offset[f]];
          for (site_count=*psize, poffset=*pcluster; site_count>0; site_count--, poffset++) {
	      rho*=spin[offset[f]+(*poffset)];
          }
	    pcur_rho[*pwhich_cluster]+=-2.*(Real)rho/rcluster_mult_per_atom[*pwhich_cluster]/rspin_size;
        }
        nomoffset=((cell[f][0]*supercell(1) + cell[f][1])*supercell(2) + cell[f][2])*site_in_cell + incell[f];
        cur_disorder_param+=(Real)(1-2*spin_changed[nomoffset])/rspin_size;
        spin_changed[nomoffset]^=1;
      }
      cur_energy+=denergy/rspin_size;
      extension_update_spin_flip(cell[1],incell[1],spin[offset[1]]);
    }
    else {
      for (mcellscan[0]=bmcell[0][0]; mcellscan[0]<total_box(0); mcellscan[0]+=supercell(0)) {
        for (mcellscan[1]=bmcell[0][1]; mcellscan[1]<total_box(1); mcellscan[1]+=supercell(1)) {
          for (mcellscan[2]=bmcell[0][2]; mcellscan[2]<total_box(2); mcellscan[2]+=supercell(2)) {
            spin[((mcellscan[0]*total_box(1) + mcellscan[1])*total_box(2) + mcellscan[2])*site_in_cell + incell[0]]*=-1;
          }
        }
      }
      extension_undo_spin_flip();
    }
}
*/

void MultiMonteCarlo::run(int mc_passes, int mode) {
  int maxn=mc_passes*supercell(0)*supercell(1)*supercell(2)*site_in_cell;
  if (mode==1) {
    for (int n=0; n<maxn; n++) {
      spin_flip();
    }
  }
  else {
    for (int n=0; n<maxn; n++) {
      //spin_double_flip();
    }
  }
}

void MultiMonteCarlo::view(const Array<Arrayint> &labellookup, const Array<AutoString> &atom_label, ofstream &file, const rMatrix3d &axes) {
  for (int i=0; i<3; i++) {
    file << axes.get_column(i) << endl;
  }
  for (int i=0; i<3; i++) {
    file << (!axes)*(supercell(i)*lattice.cell.get_column(i)) << endl;
  }
  rMatrix3d iaxes=!axes;
  MultiDimIterator<iVector3d> cur_cell(supercell);
  for ( ; cur_cell; cur_cell++) {
    iVector3d m_cur_cell=(iVector3d &)cur_cell+margin;
    for (int s=0; s<site_in_cell; s++) {
      int moffset=((m_cur_cell(0)*total_box(1) + m_cur_cell(1))*total_box(2) + m_cur_cell(2))*site_in_cell + s;
      file << iaxes*(lattice.cell*to_real(cur_cell)+lattice.atom_pos(s)) << " " << atom_label(labellookup(lattice.atom_type(s))(spin[moffset])) << endl;
    }
  }
}

void MultiMonteCarlo::get_thermo_data(Array<Real> *pdata) {
  pdata->resize(2+nb_point+total_clusters);
  (*pdata)(0)=get_cur_energy();
  (*pdata)(1)=get_cur_disorder_param();
  int i=2;
  Array<Real> conc;
  conc=get_cur_concentration();
  for (int j=0; j<conc.get_size(); j++) {
    (*pdata)(i)=conc(j);
    i++;
  }
  Array<Real> rho;
  rho=get_cur_corr();
  for (int j=0; j<rho.get_size(); j++) {
    (*pdata)(i)=rho(j);
    i++;
  }
}

void run_mc(GenericAccumulator *accum, MultiMonteCarlo *pmc, int mode) {
  Array<Real> data;
  do {
    pmc->run(1,mode);
    pmc->get_thermo_data(&data);
  } while (accum->new_data(data));
}
